from easilyb.commands import run_command_ex1
import logging
logger = logging.getLogger(__name__)


def identify(filepath):
    try:
        return_code, output = run_command_ex1(["identify", filepath])
        if return_code != 0:
            raise Exception("Non-Zero return code: %d" % return_code)
        output = output.decode(errors="replace")
        splits = output.split()
        return splits[1:4]
    except:
        logging.error("Error while running and parsing identify command for file: %s", filepath, exc_info=True)


def crop(input_filepath, output_filepath, position):
    try:
        return_code, output = run_command_ex1(["convert", input_filepath, "-crop", position, output_filepath])
        if return_code != 0:
            raise Exception("Non-Zero return code: %d" % return_code)
    except:
        logging.error("Error while running crop command for (input_filepath=%s, output_filepath=%s, position=%s)",
                      input_filepath, output_filepath, position, exc_info=True)


from threading import Lock
import os
import easilyb.json_serialize as jsons
import json
import yaml
from collections import MutableMapping

TREE_DB_META_TEMPLATE = {}


def _load_json(path, driver=jsons):
    with open(path) as fin:
        return driver.load(fin)


def _save_json(data, path, indent=None, driver=jsons):
    with open(path, 'w') as fou:
        if driver == yaml:
            driver.dump(data, fou, indent=indent, default_flow_style=False)
        else:
            driver.dump(data, fou, indent=indent)


class JsonMinConnexion:
    """Minimalistic Json Database Connexion class."""

    def __init__(self, path, create=True, template=None, template_file=None, indent=3, readonly=False, driver="jsons",
                 load=True):
        """JsonMinDb constructor.

        :param path: json file path.
        :type path: str.
        :param create: Create if json file does not exist.
        :type create: bool.
        :param template: if create is True, create from template.
        :type template: dict.
        :param template_file: if create is True and template is None, create from template_file.
        :type template_file: str.
        :param indent: if create is True, create from template.
        :type indent: int.
        :raises: ValueError

        """
        self.path = path
        self.indent = indent
        self.readonly = readonly
        self.lock = Lock()
        self.db = None
        if driver == "json":
            self.driver = json
        elif driver == "yaml":
            self.driver = yaml
        elif driver == "jsons" or driver == "json_serialize":
            self.driver = jsons
        else:
            raise ValueError("Unkown driver: %s!" % driver)
        if not os.path.isfile(path):
            if create:
                if template:
                    _template = template
                elif template_file:
                    _template = _load_json(template_file, driver=self.driver)
                else:
                    _template = {}

                _save_json(_template, path, indent=indent, driver=self.driver)
                if load:
                    self.db = _load_json(path, driver=self.driver)
            else:
                raise ValueError("Database file doesn't exist: " + path)
        if load:
            self.db = _load_json(path, driver=self.driver)

    def __contains__(self, k):
        self._load_if_not_loaded()
        return self.db.__contains__(k)

    def __delitem__(self, k):
        self._load_if_not_loaded()
        return self.db.__delitem__(k)

    def __getitem__(self, k):
        self._load_if_not_loaded()
        return self.db.__getitem__(k)

    def __iter__(self):
        self._load_if_not_loaded()
        return self.db.__iter__()

    def __len__(self):
        self._load_if_not_loaded()
        return self.db.__len__()

    def __setitem__(self, k, o):
        self._load_if_not_loaded()
        return self.db.__setitem__(k, o)

    def items(self):
        self._load_if_not_loaded()
        return self.db.items()

    def keys(self):
        self._load_if_not_loaded()
        return self.db.keys()

    def update(self):
        self._load_if_not_loaded()
        return self.db.update()

    def values(self):
        self._load_if_not_loaded()
        return self.db.values()

    def has_key(self):
        self._load_if_not_loaded()
        try:
            return self.db.has_key()
        except:
            pass

    def _load_if_not_loaded(self):
        with self.lock:
            if self.db is None:
                self.db = _load_json(self.path, driver=self.driver)

    def save(self):
        """updates database persistance file in the disk. """
        if self.readonly:
            raise Exception("Read Only Access!")
        if self.db is None:
            raise Exception("Database is not loaded")
        _save_json(self.db, self.path, indent=self.indent, driver=self.driver)

    def reload(self):
        """reload database from disk."""
        with self.lock:
            self.db = _load_json(self.path, driver=self.driver)
    
    def unload(self):
        with self.lock:
            self.db = None

    def __str__(self):
        return "<kutils.json_min_db.JsonMinConnexion instance %s>" % self.db.__str__()

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type is None:
                self.save()
        finally:
            self.lock.release()


class JsonTreeDataBase(MutableMapping):
    def __init__(self, path, create=True, indent=3, readonly=False, driver="jsons"):
        self.path = path
        self.indent = indent
        self.readonly = readonly
        self.lock = Lock()
        self.driver = driver
        self._meta = None
        self._meta_path = os.path.join(self.path, '_meta.'+ self.driver)

        if not os.path.isdir(path):
            if create:
                try:
                    os.mkdir(path)
                    self._meta = JsonMinConnexion(self._meta_path, create=create, template=TREE_DB_META_TEMPLATE,
                                                  indent=indent, readonly=readonly, driver=driver)

                except:
                    # Remove db directory if it was created

                    if os.path.isdir(self.path):
                        try: os.rmdir(self.path)
                        except: pass
                    raise
                finally:
                    pass
            else:
                raise ValueError("Database file doesn't exist: " + path)
        else:
            self._meta = JsonMinConnexion(self._meta_path, create=False, indent=indent, readonly=readonly, driver=driver)


    def __getitem__(self, item):
        pass

    def __setitem__(self, key, value):
        raise Exception('Unimplemented')

    def __delitem__(self, key):
        pass

    def __iter__(self):
        raise Exception('Unimplemented')

    def __len__(self):
        raise Exception('Unimplemented')



from easilyb.commands import run_command_ex1
import logging
logger = logging.getLogger(__name__)


def get_mouse_location():
    try:
        return_code, output = run_command_ex1(['xdotool', 'getmouselocation', '--shell'])
        if return_code != 0:
            return None
        return {k.split('=')[0]:k.split('=')[1] for k in output.decode(errors='replace').split('\n') if '=' in k}
    except:
        logger.error('Error getting mouse location', exc_info=True)
        return None



from easilyb.commands import run_command_ex1
import logging
logger = logging.getLogger(__name__)


def get_active_window_id():
    try:
        return_code, output = run_command_ex1(['xprop', '-root', '_NET_ACTIVE_WINDOW'])
        if return_code != 0:
            return None
        return output.decode(errors='replace').split()[4][:-1]
    except:
        logger.error('Error getting active window id', exc_info=True)
        return None


def get_window_title(window_id):
    try:
        return_code, output = run_command_ex1(['xprop', '-id', window_id, '_NET_WM_NAME'])
        if return_code != 0:
            return None
        return output.decode(errors='replace').split('=')[1].strip()[1:-1]
    except:
        logger.error('Error getting window title', exc_info=True)
        return None


def get_window_pid(window_id):
    try:
        return_code, output = run_command_ex1(['xprop', '-id', window_id, '_NET_WM_PID'])
        if return_code != 0:
            return None
        return int(output.decode(errors='replace').split('=')[1].strip())
    except:
        logger.error('Error getting window pid', exc_info=True)
        return None


import threading
import queue


def threaded(func, args=None, kwargs=None, daemon=False):
    args = args or ()
    kwargs = kwargs or {}

    def wrapped(q, *a, **kw):
        ret = func(*a, **kw)
        q.put(ret)

    q = queue.Queue()
    if isinstance(args, tuple):
        args = (q,)+args
    else:
        args = [q,] + args
    t = threading.Thread(target=wrapped, args=args, kwargs=kwargs)
    t.setDaemon(daemon)
    t.start()
    t.result_queue = q
    return t



import os
from easilyb.commands import get_command_output, run_command, run_command_ex1

import logging

logger = logging.getLogger(__name__)

_PS_TO_INT = ['PID', 'PPID', 'RSS']


def meminfo():
    total, free = 0, 0
    with open('/proc/meminfo') as fin:
        for line in fin.readlines():
            sline = line.split()
            if str(sline[0]) == 'MemTotal:':
                total = int(sline[1])
            elif str(sline[0]) == 'MemFree:':
                free = int(sline[1])
    return total, free


def get_processes():
    output, err = get_command_output(['ps', '-Ao', 'pid,ppid,uname,rss,comm=PROG,command=CMDLINE'])
    lines = output.split('\n')
    header = lines[0].split()
    processes = [line.split(None, len(header) - 1) for line in lines[1:] if len(line) > 0]
    processes2 = list()
    for processe in processes:
        processes2.append(
            {header[j]: (int(processe[j]) if header[j] in _PS_TO_INT else processe[j]) for j in range(len(header))})
    return processes2


def get_process(pid, processes=None):
    if processes is None: processes = get_processes()
    for process in processes:
        if process['PID'] == pid:
            return process
    return None


def get_process_children(pid=None, recursive=False, processes=None):
    childs = list()
    if pid is None:
        pid = os.getpid()
    if processes is None: processes = get_processes()
    for p in processes:
        if int(p['PPID']) == pid:
            childs.append(p)
            if recursive:
                childs.extend(get_process_children(pid=p['PPID'], recursive=False, processes=processes))
    return childs


def get_open_ports():
    return_code, output = run_command(["netstat", "-lnt"])
    ret = list()
    if return_code != 0:
        logger.error("netstat returned a non zero code: %d", return_code)
    else:
        lines = output.split('\n')
        for line in lines:
            if 'LISTEN' in line and not 'tcp6' in line:
                try:
                    parts = line.split()
                    host, port = parts[3].split(':')
                    port = int(port)
                    if not port in ret:
                        ret.append(port)
                except Exception as e:
                    logger.warning("exception parsing a line in netstat output: %s (line=%s)" % (str(e), line),
                                   exc_info=True)
    return ret


def memfree():
    try:
        return_code, output = run_command_ex1(["free", "-b"])
        if return_code != 0:
            raise Exception("Non-Zero return code: %d" % return_code)
        splits = output.decode(errors="ignore").split('\n')[1].split()
        total, available = int(splits[1]), int(splits[6])
        used = total - available
        return total, used, available
    except:
        logging.error("Error while running free command", exc_info=True)
        raise


def disk_usage(disk="/"):
    try:
        return_code, output = run_command_ex1(["df", "-B1", disk])
        if return_code != 0:
            raise Exception("Non-Zero return code: %d" % return_code)
        total, used, available = output.decode(errors="ignore").split('\n')[1].split()[1:4]
        total, available = int(total), int(available)
        used = total - available
        return total, used, available
    except:
        logging.error("Error while running df command on drive:%s", disk, exc_info=True)
        raise


def directory_usage(dir="./"):
    try:
        return_code, output = run_command_ex1(["du", "-b", "-d0", dir])
        if return_code != 0 and return_code != 1:
            raise Exception("Non-Zero return code: %d" % return_code)
        usage = int(output.decode(errors="ignore").split("\n")[-2].split()[0])
        return usage
    except:
        logging.error("Error while running du command on directory:%s", dir, exc_info=True)
        raise


def convert_bytes_to_human(b, dec=1):
    if dec > 10: dec = 10
    if dec == 0:
        if b < 1024:
            return str(b)
        elif b < 1048576:
            return str(b // 1204) + "K"
        elif b < 1073741824:
            return str(b // 1048576) + "M"
        elif b < 1099511627776:
            return str(b // 1073741824) + "G"
        else:
            return str(b // 1099511627776) + "T"
    if b < 1024:
        return str(b)
    elif b < 1048576:
        return str(int(b / 1204 * pow(10, dec)) / pow(10, dec)) + "K"
    elif b < 1073741824:
        return str(int(b / 1048576 * pow(10, dec)) / pow(10, dec)) + "M"
    elif b < 1099511627776:
        return str(int(b / 1073741824 * pow(10, dec)) / pow(10, dec)) + "G"
    else:
        return str(int(b / 1099511627776 * pow(10, dec)) / pow(10, dec)) + "T"


MULTIPLIER = {"b":1, "k": 1024, "m": 1048576, "g": 1073741824, "t": 1099511627776}


def convert_bytes_from_human(h):
    try:
        return float(h)
    except:
        dec, mult = float(h[0:-1]), h[-1]
        return int(dec * MULTIPLIER[mult.lower()])



from easilyb.commands import run_command_ex1
import logging
logger = logging.getLogger(__name__)


def get_process_cmd(pid):
    try:
        return_code, output = run_command_ex1(['ps', '--no-headers', '-fw', '--pid', str(pid)])
        if return_code != 0:
            return None
        return output.decode(errors='replace').split()[7].strip()
    except:
        logger.error('Error getting process cmd', exc_info=True)
        return None


import sys
import logging
logger = logging.getLogger(__name__)


def get_active_window():
    if sys.platform.startswith('linux'):
        try:
            import easilyb.os_ops.linux.xprop as _xprop
            import easilyb.os_ops.linux.ps as _ps
            window = dict()
            wid = _xprop.get_active_window_id()
            window['pid'] = _xprop.get_window_pid(wid)
            window['title'] = _xprop.get_window_title(wid)
            window['cmd'] = _ps.get_process_cmd(window['pid'])
            return window
        except:
            logger.error('Error getting active window for Linux platform', exc_info=True)
            raise Exception('Error getting active window for Linux platform')
    else:
        logger.error('platform %s is not supported', sys.platform)
        raise Exception('platform %s is not supported' % sys.platform)


def get_mouse_location():
    if sys.platform.startswith('linux'):
        import easilyb.os_ops.linux.xdotool as _xdotool
        try:
            ml = _xdotool.get_mouse_location()
            return int(ml['X']), int(ml['Y'])
        except:
            logger.error('Error getting mouse location for Linux platform', exc_info=True)
            raise Exception('Error getting mouse location for Linux platform')
    else:
        logger.error('platform %s is not supported', sys.platform)
        raise Exception('platform %s is not supported' % sys.platform)


from easilyb.time_ops._time_ops import epoch_to_pathfriendly, epoch_to_iso8601, format_utc_time, format_utc_date, \
    parse_utc_date, parse_utc_time, UTC_ISO8601_TIME_FORMAT, UTC_ISO8601_DATE_FORMAT


from ipaddress import ip_address, ip_network, _BaseAddress
import logging
logger = logging.getLogger(__name__)


class _BaseIPDataClass:
    def __str__(self):
        return str(self.__class__) + "." + self.ID

    def __repr__(self):
        return "'" + self.__str__() + "'"

    def __hash__(self):
        return hash(self.__class__.__name__ + "." + self.ID)

    def __eq__(self, other):
        try:
            return self.__class__ == other.__class__ and self.ID == other.ID
        except:
            return False


class IP(_BaseIPDataClass):
    def __init__(self, ip):
        if not isinstance(ip, _BaseAddress):
            self._ip = ip_address(ip)
        else:
            self._ip = ip
        self.version = self._ip.version
        self.decimal = int(self._ip)
        self.compressed = self._ip.compressed
        self.exploded = self._ip.exploded
        self.ID = self.exploded

    def __int__(self):
        return self.decimal

    def __lt__(self, other):
        if not isinstance(other, IP):
            other = IP(other)
        return self._ip.__lt__(other._ip)

    def __le__(self, other):
        if not isinstance(other, IP):
            other = IP(other)
        return self == other or self._ip.__lt__(other._ip)

    def __add__(self, other):
        return IP(self.decimal + other)

    def __sub__(self, other):
        return IP(self.decimal - other)


class Pool(_BaseIPDataClass):
    @staticmethod
    def from_netrange(pool):
        pool = pool.replace(' ', '')
        lower, upper = None, None
        if '/' in pool:
            try:
                net = ip_network(pool)
                lower, upper = ip_address(int(net.network_address) + 1), ip_address(int(net.broadcast_address) - 1)
            except Exception as e:
                raise ValueError("Bad CIDR pool: " + pool) from e
        elif '-' in pool:
            try:
                lower, upper = pool.split('-')
            except Exception as e:
                raise ValueError("Bad range pool: " + pool) from e
        else:
            try:
                single_ip_pool = ip_address(pool)
                lower, upper = single_ip_pool, single_ip_pool
            except:
                raise ValueError("Bad pool: " + pool)
        return Pool(lower, upper)

    def __init__(self, lower, upper):
        if not isinstance(lower, IP):
            self.lower = IP(lower)
        else:
            self.lower = lower
        if not isinstance(upper, IP):
            self.upper = IP(upper)
        else:
            self.upper = upper
        if self.lower.version != self.upper.version:
            raise ValueError("IPs are not the same version")
        self.version = self.lower.version
        self.normalized = self.lower.exploded + ' - ' + self.upper.exploded
        self.count = (int(self.upper) - int(self.lower)) + 1
        self.ID = self.normalized

    def __contains__(self, ip):
        if not isinstance(ip, IP):
            ip = IP(ip)
        if ip.version != self.version:
            return False
        return self.lower.decimal <= ip.decimal <= self.upper.decimal

    def __str__(self):
        return self.normalized
